@@ -2,23 +2,23 @@
  * msg.c - routines for error messages.
  */
 
-/* 
- * Copyright (C) 1986, 1988, 1989, 1991-2001, 2003, 2010
+/*
+ * Copyright (C) 1986, 1988, 1989, 1991-2001, 2003, 2010-2013, 2017, 2018,
  * the Free Software Foundation, Inc.
- * 
+ *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
- * 
+ *
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2, or (at your option)
  * any later version.
- * 
+ *
  * GAWK is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
@@ -33,28 +33,40 @@ static const char *srcfile = NULL;
 static int srcline;
 
 jmp_buf fatal_tag;
-int fatal_tag_valid = FALSE;
+int fatal_tag_valid = 0;
 
 /* err --- print an error message with source line and file and record */
 
 /* VARARGS2 */
 void
-err(const char *s, const char *emsg, va_list argp)
+err(bool isfatal, const char *s, const char *emsg, va_list argp)
 {
 	char *file;
 	const char *me;
 
+	static bool first = true;
+	static bool add_src_info = false;
+	static long lineno_val = 0;	// Easter Egg
+
+	if (first) {
+		first = false;
+		add_src_info = (getenv("GAWK_MSG_SRC") != NULL);
+		if (! do_traditional) {
+			NODE *n = lookup("LINENO");
+
+			if (n != NULL && n->type == Node_var)
+				lineno_val = get_number_d(n->var_value);
+		}
+	}
+
 	(void) fflush(output_fp);
 	me = myname;
-	if (strncmp(me, "dgawk", 5) == 0)
-		me = &myname[1];
 	(void) fprintf(stderr, "%s: ", me);
-#ifdef GAWKDEBUG
-	if (srcfile != NULL) {
+
+	if (srcfile != NULL && add_src_info) {
 		fprintf(stderr, "%s:%d:", srcfile, srcline);
 		srcfile = NULL;
 	}
-#endif /* GAWKDEBUG */
 
 	if (sourceline > 0) {
 		if (source != NULL)
@@ -62,19 +74,46 @@ err(const char *s, const char *emsg, va_list argp)
 		else
 			(void) fprintf(stderr, _("cmd. line:"));
 
-		(void) fprintf(stderr, "%d: ", sourceline);
+		(void) fprintf(stderr, "%ld: ", sourceline + lineno_val);
 	}
+
+#ifdef HAVE_MPFR
+	if (FNR_node && is_mpg_number(FNR_node->var_value)) {
+		NODE *val;
+		val = mpg_update_var(FNR_node);
+		assert((val->flags & MPZN) != 0);
+		if (mpz_sgn(val->mpg_i) > 0) {
+			int len = FILENAME_node->var_value->stlen;
+			file = FILENAME_node->var_value->stptr;
+			(void) putc('(', stderr);
+			if (file)
+				(void) fprintf(stderr, "FILENAME=%.*s ", len, file);
+			(void) mpfr_fprintf(stderr, "FNR=%Zd) ", val->mpg_i);
+		}
+	} else
+#endif
 	if (FNR > 0) {
+		int len = FILENAME_node->var_value->stlen;
 		file = FILENAME_node->var_value->stptr;
 		(void) putc('(', stderr);
 		if (file)
-			(void) fprintf(stderr, "FILENAME=%s ", file);
+			(void) fprintf(stderr, "FILENAME=%.*s ", len, file);
 		(void) fprintf(stderr, "FNR=%ld) ", FNR);
 	}
+
 	(void) fprintf(stderr, "%s", s);
 	vfprintf(stderr, emsg, argp);
 	(void) fprintf(stderr, "\n");
 	(void) fflush(stderr);
+
+	if (isfatal) {
+#ifdef GAWKDEBUG
+		// GLIBC 2.27 doesn't necessarily flush on abort. Sigh.
+		fflush(NULL);
+		abort();
+#endif
+		gawk_exit(EXIT_FATAL);
+	}
 }
 
 /* msg --- take a varargs error message and print it */
@@ -84,18 +123,18 @@ msg(const char *mesg, ...)
 {
 	va_list args;
 	va_start(args, mesg);
-	err("", mesg, args);
+	err(false, "", mesg, args);
 	va_end(args);
 }
 
-/* warning --- print a warning message */
+/* r_warning --- print a warning message */
 
 void
-warning(const char *mesg, ...)
+r_warning(const char *mesg, ...)
 {
 	va_list args;
 	va_start(args, mesg);
-	err(_("warning: "), mesg, args);
+	err(false, _("warning: "), mesg, args);
 	va_end(args);
 }
 
@@ -104,7 +143,7 @@ error(const char *mesg, ...)
 {
 	va_list args;
 	va_start(args, mesg);
-	err(_("error: "), mesg, args);
+	err(false, _("error: "), mesg, args);
 	va_end(args);
 }
 
@@ -127,12 +166,8 @@ r_fatal(const char *mesg, ...)
 {
 	va_list args;
 	va_start(args, mesg);
-	err(_("fatal: "), mesg, args);
+	err(true, _("fatal: "), mesg, args);
 	va_end(args);
-#ifdef GAWKDEBUG
-	abort();
-#endif
-	gawk_exit(EXIT_FATAL);
 }
 
 /* gawk_exit --- longjmp out if necessary */
@@ -144,5 +179,20 @@ gawk_exit(int status)
 		exit_val = status;
 		longjmp(fatal_tag, 1);
 	}
+
+	final_exit(status);
+}
+
+/* final_exit --- run extension exit handlers and exit */
+
+void
+final_exit(int status)
+{
+	/* run any extension exit handlers */
+	run_ext_exit_handlers(status);
+
+	/* we could close_io() here */
+	close_extensions();
+
 	exit(status);
 }
